# AES import party!
# pycryptodome and PyCryptodomex's implementations of AES are the same,
# so we try to import one of these twos.
import argparse
import io
import zlib
from datetime import date
from re import findall

import logging
from time import sleep

from wa_crypt_tools.lib.constants import C
from wa_crypt_tools.lib.key.keyfactory import KeyFactory
from wa_crypt_tools.lib.logformat import CustomFormatter
from wa_crypt_tools.lib.utils import test_decompression

log = logging.getLogger(__name__)

# AES import party!
# pycryptodome and PyCryptodomex's implementations of AES are the same,
# so we try to import one of these twos.
try:
    # pycryptodomex
    from Cryptodome.Cipher import AES
except ModuleNotFoundError:
    try:
        # pycryptodome
        # noinspection PyUnresolvedReferences
        from Crypto.Cipher import AES

        if not hasattr(AES, 'MODE_GCM'):
            # pycrypto
            raise ModuleNotFoundError("You installed pycrypto and not pycryptodome(x). "
                                      "Pycrypto is old, deprecated and not supported. \n"
                                      "Run: python -m pip uninstall pycrypto\n"
                                      "And: python -m pip install pycryptodomex\n"
                                      "Or:  python -m pip install pycryptodome")
    except ModuleNotFoundError:
        # crypto (or nothing)
        raise ModuleNotFoundError("You need pycryptodome(x) to run these scripts!\n"
                                  "python -m pip install pycryptodome\n"
                                  "Or: python -m pip install pycryptodome\n"
                                  "You can also remove \"crypto\" if you have it installed\n"
                                  "python -m pip uninstall crypto")


def oscillate(n: int, n_min: int, n_max: int):
    """Yields n, n-1, n+1, n-2, n+2..., with constraints:
    - n is in [min, max]
    - n is never negative
    Reverts to range() when n touches min or max. Example:
    oscillate(8, 2, 10) => 8, 7, 9, 6, 10, 5, 4, 3, 2
    """

    if n_min < 0:
        n_min = 0

    i = n
    c = 1

    # First phase (n, n-1, n+1...)
    while True:

        if i == n_max:
            break
        yield i
        i -= c
        c += 1

        if i == 0 or i == n_min:
            break
        yield i
        i += c
        c += 1

    # Second phase (range of remaining numbers)
    # n != i/2 fixes a bug where we would yield min and max two times if n == (max-min)/2
    if i == n_min and n != i / 2:

        yield i
        i += c
        for j in range(i, n_max + 1):
            yield j

    if i == n_max and n != i / 2:

        yield n_max
        i -= c
        for j in range(i, n_min - 1, -1):
            yield j


def find_data_offset(header: bytes, iv_offset: int, key: bytes, starting_data_offset: int) -> int:
    """Tries to find the offset in which the encrypted data starts.
    Returns the offset or -1 if the offset is not found.
    Only works with ZLIB stream, not with ZIP file."""

    iv = header[iv_offset:iv_offset + 16]

    # oscillate ensures we try the closest values to the default value first.
    for i in oscillate(n=starting_data_offset, n_min=iv_offset + len(iv), n_max=C.HEADER_SIZE - 128):

        cipher = AES.new(key, AES.MODE_GCM, iv)

        # We only decrypt the first two bytes.
        test_bytes = cipher.decrypt(header[i:i + 2])

        for zheader in C.ZLIB_HEADERS:

            if test_bytes == zheader:
                # We found a match, but this might also happen by chance.
                # Let's run another test by decrypting some hundreds of bytes.
                # We need to reinitialize the cipher everytime as it has an internal status.
                cipher = AES.new(key, AES.MODE_GCM, iv)
                decrypted = cipher.decrypt(header[i:])
                if test_decompression(decrypted):
                    return i
    return -1


def guess_offsets(key: bytes, encrypted: io.BufferedReader, def_iv_offset: int,
                  def_data_offset: int):
    """Gets the IV, shifts the stream to the beginning of the encrypted data and returns the cipher.
    It does so by guessing the offset."""

    # Assign variables to suppress warnings
    db_header, data_offset, iv_offset = None, None, None

    # Restart the file stream
    encrypted.seek(0)

    db_header = encrypted.read(C.HEADER_SIZE)
    if len(db_header) < C.HEADER_SIZE:
        log.fatal("The encrypted database is too small.\n    "
                  "Did you swap the keyfile and the encrypted database file by mistake?")

    try:
        if db_header[:15].decode('ascii') == 'SQLite format 3':
            log.error("The database file is not encrypted.\n    "
                      "Did you swap the input and the output files by mistake?")
    except ValueError:
        pass

    # Finding WhatsApp's version is nice
    version = findall(b"\\d(?:\\.\\d{1,3}){3}", db_header)
    if len(version) != 1:
        log.info('WhatsApp version not found (Crypt12?)')
    else:
        log.debug("WhatsApp version: {}".format(version[0].decode('ascii')))

    # Determine IV offset and data offset.
    for iv_offset in oscillate(n=def_iv_offset, n_min=0, n_max=C.HEADER_SIZE - 128):
        data_offset = find_data_offset(db_header, iv_offset, key, def_data_offset)
        if data_offset != -1:
            log.info("Offsets guessed (IV: {}, data: {}).".format(iv_offset, data_offset))
            if iv_offset != def_iv_offset or data_offset != def_data_offset:
                log.info("Next time, use -ivo {} -do {} for guess-free decryption".format(iv_offset, data_offset))
            break
    if data_offset == -1:
        return None

    iv = db_header[iv_offset:iv_offset + 16]

    encrypted.seek(data_offset)

    return AES.new(key, AES.MODE_GCM, iv)


def parsecmdline() -> argparse.Namespace:
    """Sets up the argument parser"""
    parser = argparse.ArgumentParser(description='Decrypts WhatsApp backup files'
                                                 ' encrypted with crypt12, 14 or 15')
    parser.add_argument('keyfile', nargs='?', type=str, default="encrypted_backup.key",
                        help='The WhatsApp encrypted_backup key file or the hex encoded key. '
                             'Default: encrypted_backup.key')
    parser.add_argument('encrypted', nargs='?', type=argparse.FileType('rb'), default="msgstore.db.crypt15",
                        help='The encrypted crypt12, 14 or 15 file. Default: msgstore.db.crypt15')
    parser.add_argument('decrypted', nargs='?', type=argparse.FileType('wb'), default="msgstore.db",
                        help='The decrypted output file. Default: msgstore.db')
    parser.add_argument('-ivo', '--iv-offset', type=int, default=C.DEFAULT_IV_OFFSET,
                        help='The default offset of the IV in the encrypted file. '
                             'Default: {}'.format(C.DEFAULT_IV_OFFSET))
    parser.add_argument('-do', '--data-offset', type=int, default=C.DEFAULT_DATA_OFFSET,
                        help='The default offset of the encrypted data in the encrypted file. '
                             'Default: {}'.format(C.DEFAULT_DATA_OFFSET))
    parser.add_argument('-v', '--verbose', action='store_true', help='Prints all offsets and messages')

    return parser.parse_args()


def decrypt(cipher, encrypted, decrypted):
    """Does the actual decryption."""

    z_obj = zlib.decompressobj()

    if cipher is None:
        log.fatal("Could not create a decryption cipher")

    try:

        try:
            encrypted_data = encrypted.read()
            # Crypt12 moment: the last 4 bytes are --xx, where xx
            # are the last 2 numbers of the jid (user's phone number).
            # We need to remove them.

            try:
                output_decrypted: bytearray = cipher.decrypt(encrypted_data)
            except ValueError as e:
                log.fatal("Decryption failed: {}."
                          "\n    This probably means your backup is corrupted.".format(e))
                # Dead code to make pycharm warning go away
                exit(1)

            try:
                output_file = z_obj.decompress(output_decrypted)
                if not z_obj.eof:
                    log.error("The encrypted database file is truncated (damaged).")
            except zlib.error:
                output_file = output_decrypted
                if test_decompression(output_file[:io.DEFAULT_BUFFER_SIZE]):
                    log.info("Decrypted data is a ZIP file that I will not decompress automatically.")
                else:
                    log.error("I can't recognize decrypted data. Decryption not successful.\n    "
                              "The key probably does not match with the encrypted file.\n    "
                              "Or the backup is simply empty. (check with --force)")

            decrypted.write(output_file)

        except MemoryError:
            log.fatal("Out of RAM, please use -nm.")

        decrypted.flush()

    except OSError as e:
        log.fatal("I/O error: {}".format(e))

    finally:
        decrypted.close()
        encrypted.close()


def main():
    args = parsecmdline()

    # set wa_crypt_tools l to debug
    log.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch.setFormatter(CustomFormatter())
    log.addHandler(ch)
    if not (0 < args.data_offset < C.HEADER_SIZE - 128):
        log.fatal("The data offset must be between 1 and {}".format(C.HEADER_SIZE - 129))
    if not (0 < args.iv_offset < C.HEADER_SIZE - 128):
        log.fatal("The IV offset must be between 1 and {}".format(C.HEADER_SIZE - 129))
    # Get the decryption key from the key file or the hex encoded string.
    key = KeyFactory.new(args.keyfile)
    log.debug(str(key))

    cipher = guess_offsets(key=key.get(), encrypted=args.encrypted,
                           def_iv_offset=args.iv_offset, def_data_offset=args.data_offset)

    decrypt(cipher, args.encrypted, args.decrypted)

    if date.today().day == 1 and date.today().month == 4:
        log.info("Done. Uploading messages to the developer's server...")
        sleep(0.5)
        log.info("Uploaded. The developer will now read and publish your messages!")
    else:
        log.info("Done")


if __name__ == '__main__':
    main()



from hashlib import sha256
from os import urandom
from pathlib import Path

from javaobj import JavaObjectMarshaller

from wa_crypt_tools.lib.key.key import Key
from wa_crypt_tools.lib.utils import create_jba

import logging

log = logging.getLogger(__name__)


class Key14(Key):
    # These constants are only used with crypt12/14 keys.
    __SUPPORTED_CIPHER_VERSION = b'\x00\x01'
    __SUPPORTED_KEY_VERSIONS = [b'\x01', b'\x02', b'\x03']

    def __init__(self, keyarray: bytes = None,
                 cipher_version: bytes = None, key_version: bytes = None,
                 serversalt: bytes = None, googleid: bytes = None, hashedgoogleid: bytes = None,
                 iv: bytes = None, key: bytes = None):
        """Extracts the fields from a crypt14 loaded key file."""
        # key file format and encoding explanation:
        # The key file is actually a serialized byte[] object.

        # After deserialization, we will have a byte[] object that we have to split in:
        # 1) The cipher version (2 bytes). Known values are 0x0000 and 0x0001. So far we only support the latter.
        # SUPPORTED_CIPHER_VERSION = b'\x00\x01'
        # 2) The key version (1 byte). All the known versions are supported.
        # SUPPORTED_KEY_VERSIONS = [b'\x01', b'\x02', b'\x03']
        # Looks like nothing actually changes between the versions.
        # 3) Server salt (32 bytes)
        # 4) googleIdSalt (unused?) (16 bytes)
        # 5) hashedGoogleID (The SHA-256 hash of googleIdSalt) (32 bytes)
        # 6) encryption IV (zeroed out, as it is read from the database) (16 bytes)
        # 7) cipherKey (The actual AES-256 decryption key) (32 bytes)

        if keyarray is None:
            # Randomly generated key or with supplied parameters
            if cipher_version is None:
                self.__cipher_version = self.__SUPPORTED_CIPHER_VERSION
            else:
                if cipher_version != self.__SUPPORTED_CIPHER_VERSION:
                    log.error("Invalid cipher version: {}".format(cipher_version.hex()))
                self.__cipher_version = cipher_version
            if key_version is None:
                self.__key_version = self.__SUPPORTED_KEY_VERSIONS[-1]
            else:
                if key_version not in self.__SUPPORTED_KEY_VERSIONS:
                    log.error("Invalid key version: {}".format(key_version.hex()))
                self.__key_version = key_version
            if serversalt is None:
                self.__serversalt = urandom(32)
            else:
                if len(serversalt) != 32:
                    log.error("Invalid server salt length: {}".format(serversalt.hex()))
                self.__serversalt = serversalt
            if googleid is None:
                self.__googleid = urandom(16)
            else:
                if len(googleid) != 16:
                    log.error("Invalid google id length: {}".format(googleid.hex()))
                self.__googleid = googleid
            if hashedgoogleid is None:
                self.__hashedgoogleid = sha256(self.__googleid).digest()
            else:
                log.warning("Using supplied hashed google id")
                if len(hashedgoogleid) != 32:
                    log.error("Invalid hashed google id length: {}".format(hashedgoogleid.hex()))
                self.__hashedgoogleid = hashedgoogleid
            if iv is None:
                self.__padding = b'\x00' * 16
            else:
                if len(iv) != 16:
                    log.error("Invalid IV length: {}".format(iv.hex()))
                if iv != b'\x00' * 16:
                    log.warning("IV should be empty")
                self.__padding = iv
            if key is None:
                self.__key = urandom(32)
            else:
                if len(key) != 32:
                    log.error("Invalid key length: {}".format(key.hex()))
                self.__key = key
            return
        # Check if the keyfile has a supported cipher version
        self.__cipher_version = keyarray[:len(self.__SUPPORTED_CIPHER_VERSION)]
        if self.__SUPPORTED_CIPHER_VERSION != self.__cipher_version:
            log.error("Invalid keyfile: Unsupported cipher version {}"
                      .format(keyarray[:len(self.__SUPPORTED_CIPHER_VERSION)].hex()))
        index = len(self.__SUPPORTED_CIPHER_VERSION)

        # Check if the keyfile has a supported key version
        version_supported = False
        for v in self.__SUPPORTED_KEY_VERSIONS:
            if v == keyarray[index:index + len(self.__SUPPORTED_KEY_VERSIONS[0])]:
                version_supported = True
                self.__key_version = v
                break
        if not version_supported:
            log.error('Invalid keyfile: Unsupported key version {}'
                      .format(keyarray[index:index + len(self.__SUPPORTED_KEY_VERSIONS[0])].hex()))

        self.__serversalt = keyarray[3:35]

        # Check the SHA-256 of the salt
        self.__googleid = keyarray[35:51]
        expected_digest = sha256(self.__googleid).digest()
        actual_digest = keyarray[51:83]
        if expected_digest != actual_digest:
            log.error("Invalid keyfile: Invalid SHA-256 of salt.\n    "
                      "Expected: {}\n    Got:{}".format(expected_digest, actual_digest))

        self.__hashedgoogleid = actual_digest

        self.__padding = keyarray[83:99]

        # Check if IV is made of zeroes
        for byte in self.__padding:
            if byte:
                log.error("Invalid keyfile: IV is not zeroed out but is: {}".format(self.__padding.hex()))
                break

        self.__key = keyarray[99:]

        log.info("Crypt12/14 key loaded")

    def get(self) -> bytes:
        return self.__key

    def get_serversalt(self) -> bytes:
        return self.__serversalt

    def get_googleid(self) -> bytes:
        return self.__googleid

    def get_cipher_version(self) -> bytes:
        return self.__cipher_version

    def get_key_version(self) -> bytes:
        return self.__key_version

    def __str__(self) -> str:
        """Returns a string representation of the key"""
        try:
            string: str = "Key14("
            if self.__key is not None:
                string += "key: {}".format(self.__key.hex())
            if self.__serversalt is not None:
                string += " , serversalt: {}".format(self.__serversalt.hex())
            if self.__googleid is not None:
                string += " , googleid: {}".format(self.__googleid.hex())
            if self.__key_version is not None:
                string += " , key_version: {}".format(self.__key_version.hex())
            if self.__cipher_version is not None:
                string += " , cipher_version: {}".format(self.__cipher_version.hex())
            return string + ")"
        except Exception as e:
            return "Exception printing key: {}".format(e)

    def __repr__(self) -> str:
        # TODO
        return self.__str__()

    def dump(self) -> bytes:
        """Dumps the key to a file"""
        out: bytes = b''
        out += self.__cipher_version
        out += self.__key_version
        out += self.__serversalt
        out += self.__googleid
        out += self.__hashedgoogleid
        out += self.__padding
        out += self.__key
        return JavaObjectMarshaller().dump(create_jba(out))

    def file_dump(self, file: Path):
        with open(file, 'wb') as f:
            f.write(self.dump())



#!/usr/bin/env python
"""
This script prints info on WhatsApp's DB files.
"""

from __future__ import annotations

from wa_crypt_tools.lib.logformat import CustomFormatter
from wa_crypt_tools.lib.db.dbfactory import DatabaseFactory
from wa_crypt_tools.lib.key.keyfactory import KeyFactory

import argparse

__author__ = 'ElDavo'
__copyright__ = 'Copyright (C) 2024'
__license__ = 'GPLv3'
__status__ = 'Beta'

import logging

log = logging.getLogger(__name__)


def parsecmdline() -> argparse.Namespace:
    """Sets up the argument parser"""
    parser = argparse.ArgumentParser(description='Prints info on whatsapp crypted files')
    parser.add_argument('encrypted', nargs='?',
                        type=str,
                        default="msgstore.db.crypt15",
                        help='The encrypted crypt12, 14 or 15 file. Default: msgstore.db.crypt15')
    parser.add_argument('-k', '--key',
                        action='store_true',
                        help='tell the program that the file is a key file')
    return parser.parse_args()


def main():
    args = parsecmdline()

    # set wa_crypt_tools l to debug
    log.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(CustomFormatter())
    log.addHandler(ch)
    # also add to "wa_crypt_tools.lib" logger
    logging.getLogger("wa_crypt_tools.lib").addHandler(ch)
    logging.getLogger("wa_crypt_tools.lib").setLevel(logging.DEBUG)

    log.warning("This script is in beta stage.")

    if args.key:
        key = KeyFactory.from_file(args.encrypted)
        print(key)
        return
    try:
        DatabaseFactory.from_file(open(args.encrypted, 'rb'))
    except Exception as e:
        log.error("Error: {}".format(e))
        return
        # TODO
    # print(db)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
This script decrypts WhatsApp's DB files encrypted with Crypt12, Crypt14 or Crypt15.
"""

from __future__ import annotations

import os
from pathlib import Path

from wa_crypt_tools.lib.key.key14 import Key14
from wa_crypt_tools.lib.key.key15 import Key15
from wa_crypt_tools.lib.logformat import CustomFormatter
import argparse

__author__ = 'ElDavo'
__copyright__ = 'Copyright (C) 2023'
__license__ = 'GPLv3'
__status__ = 'Production'

import logging

lo = logging.getLogger(__name__)


def parsecmdline() -> argparse.Namespace:
    """Sets up the argument parser"""
    parser = argparse.ArgumentParser(description='Create a key or encrypted_backup.key from a hex input.'
                                                 'The only parameter a encrypted_backup.key stores is the key itself.')
    parser.add_argument('-c14', '--crypt14', action='store_true', default=False, help='Create a traditional key file.')
    parser.add_argument('-o', '--output', type=argparse.FileType('wb'),
                        help='The output file')
    parser.add_argument('-y', '--yes', action='store_true', help='Overwrite the output file if it exists.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Prints all messages')
    parser.add_argument('--hex', type=str, nargs='?', help='The hex string to convert to a key')

    parser.add_argument('-cv', '--cipher-version', type=int, help='The cipher version to use. Default: 1')
    parser.add_argument('-kv', '--key-version', type=int, help='The key version to use. Default: 3')
    parser.add_argument('-ss', '--server-salt', type=str, help='The server salt to use. Default: random')
    parser.add_argument('-gi', '--googleid', type=str, help='The google id salt to use. Default: random')

    return parser.parse_args()


def main():
    args = parsecmdline()

    # set wa_crypt_tools l to debug
    lo.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch.setFormatter(CustomFormatter())
    lo.addHandler(ch)

    if args.hex is None:
        lo.warning("Key not specified, a random key will be generated.")

    if args.output is None:
        args.output = "key" if args.crypt14 else "encrypted_backup.key"

    if args.crypt14:
        if args.cipher_version is None:
            args.cipher_version = 1
        if args.key_version is None:
            args.key_version = 3
        if args.server_salt is None:
            lo.warning("Server salt not specified, a random one will be generated.")
        if args.googleid is None:
            lo.warning("Google id not specified, a random one will be generated.")
        key: Key14 = Key14(cipher_version=args.cipher_version.to_bytes(2, "big"),
                           key_version=args.key_version.to_bytes(1, "big"),
                           serversalt=args.server_salt, googleid=args.googleid,
                           iv=None, key=None)
    else:
        if args.cipher_version is not None:
            lo.warning("Cipher version specified, but it is not used for crypt15 keys, ignoring.")
        if args.key_version is not None:
            lo.warning("Key version specified, but it is not used for crypt15 keys, ignoring.")
        if args.server_salt is not None:
            lo.warning("Server salt specified, but it is not used for crypt15 keys, ignoring.")
        if args.googleid is not None:
            lo.warning("Google id specified, but it is not used for crypt15 keys, ignoring.")
        key: Key15 = Key15(keyarray=bytes.fromhex(args.hex) if args.hex is not None else None)

    # Check if the output file exists
    output_file = Path(args.output)
    print(os.getcwd())
    if output_file.is_file() and not args.yes:
        lo.fatal("The output file already exists.")
        return

    # Write the key file
    key.file_dump(output_file)

    lo.info("Key file \"{}\" created.".format(args.output))


if __name__ == "__main__":
    main()



import hmac
from hashlib import sha256
from os import urandom
from pathlib import Path

from javaobj import JavaObjectMarshaller

from wa_crypt_tools.lib.utils import create_jba, encryptionloop

from wa_crypt_tools.lib.key.key import Key
import logging

l = logging.getLogger(__name__)


class Key15(Key):
    # This constant is only used with crypt15 keys.
    BACKUP_ENCRYPTION = b'backup encryption'

    def __init__(self, keyarray: bytes = None, key: bytes = None):
        """Extracts the key from a loaded crypt15 key file."""
        # encrypted_backup.key file format and encoding explanation:
        # The E2E key file is actually a serialized byte[] object.

        # After deserialization, we will have the root key (32 bytes).
        # The root key is further encoded with three different strings, depending on what you want to do.
        # These three ways are "backup encryption";
        # "metadata encryption" and "metadata authentication", for Google Drive E2E encrypted metadata.
        # We are only interested in the local backup encryption.

        # Why the \x01 at the end of the BACKUP_ENCRYPTION constant?
        # Whatsapp uses a nested encryption function to encrypt many times the same data.
        # The iteration counter is appended to the end of the encrypted data. However,
        # since the loop is actually executed only one time, we will only have one interaction,
        # and thus a \x01 at the end.
        # Take a look at utils/wa_hmacsha256_loop.java that is the original code.

        if keyarray is None:
            # Randomly generated key or with supplied parameters
            if key is None:
                self.__key = urandom(32)
            else:
                if len(key) != 32:
                    l.error("Invalid key length: {}".format(key.hex()))
                self.__key = key
            return

        if len(keyarray) != 32:
            l.critical("Crypt15 loader trying to load a crypt14 key")
        l.debug("Root key: {}".format(keyarray.hex()))
        # Save the root key in the class
        self.__key = keyarray

        l.info("Crypt15 / Raw key loaded")

    def get(self) -> bytes:
        """
        Returns the key used for encryption, that is not the root key.
        """
        return encryptionloop(
            first_iteration_data=self.__key,
            message=b'backup encryption',
            output_bytes=32)

    def get_root(self) -> bytes:
        """
        Returns the root key.
        """
        return self.__key

    def get_metadata_encryption(self) -> bytes:
        """
        Returns the key used for metadata encryption
        """
        return encryptionloop(
            first_iteration_data=self.__key,
            message=b'metadata encryption',
            output_bytes=32)

    def get_metadata_authentication(self) -> bytes:
        """
        Returns the key used for metadata authentication
        """
        return encryptionloop(
            first_iteration_data=self.__key,
            message=b'metadata authentication',
            output_bytes=32)

    def dump(self) -> bytes:
        """Dumps the key"""
        return JavaObjectMarshaller().dump(create_jba(self.__key))

    def file_dump(self, file: Path):
        with open(file, 'wb') as f:
            f.write(self.dump())

    def __str__(self) -> str:
        """Returns a string representation of the key"""
        try:
            string: str = "Key15("
            if self.__key is not None:
                string += "key: {}".format(self.__key.hex())
            return string + ")"
        except Exception as e:
            return "Exception printing key: {}".format(e)

    def __repr__(self) -> str:
        # TODO
        return self.__str__()



import os
import zlib

from wa_crypt_tools.lib.db.db12 import Database12
from wa_crypt_tools.lib.db.db14 import Database14
from wa_crypt_tools.lib.db.db15 import Database15
from wa_crypt_tools.lib.key.keyfactory import KeyFactory
from wa_crypt_tools.lib.props import Props
from hashlib import sha512


class TestEncryption:
    def test_encryption15(self):
        key = KeyFactory.new("tests/res/encrypted_backup.key")
        props = Props(wa_version="2.22.5.13", jid="67", features=[5, 7, 8, 13, 14, 19, 22, 25, 28, 30, 31, 32, 36, 37],
                      max_feature=37)
        db = Database15(key=key, iv=bytes.fromhex("C395EE009CF8B68AC0EA760550F6559C"))
        data = db.encrypt(
            key,
            props,
            zlib.compress(
                open("tests/res/msgstore.db", 'rb').read(),
                level=1,
            )
        )
        new_check = sha512(data).digest()
        with open("tests/res/msgstore-new.db.crypt15", 'wb') as f:
            f.write(data)
        with open("tests/res/msgstore.db.crypt15", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert new_check == orig_check
        os.remove("tests/res/msgstore-new.db.crypt15")

    def test_encryption14(self):
        key = KeyFactory.new("tests/res/key")
        props = Props(wa_version="2.22.5.13", jid="67", features=[5, 7, 8, 13, 14, 19, 22, 25, 28, 30, 31, 32, 36, 37],
                      max_feature=37)
        db = Database14(key=key, iv=bytes.fromhex("EA53CEAE36ECAB50BC331AEB62491625"))
        data = db.encrypt(
            key,
            props,
            zlib.compress(
                open("tests/res/msgstore.db", 'rb').read(),
                level=1,
            )
        )
        new_check = sha512(data).digest()
        with open("tests/res/msgstore-new.db.crypt14", 'wb') as f:
            f.write(data)
        with open("tests/res/msgstore.db.crypt14", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert new_check == orig_check
        os.remove("tests/res/msgstore-new.db.crypt14")

    def test_encryption14_noexpiry(self):
        key = KeyFactory.new("tests/res/key")
        props = Props(wa_version="2.22.5.13", jid="67", features=None)
        db = Database14(key=key, iv=bytes.fromhex("EA53CEAE36ECAB50BC331AEB62491625"))
        data = db.encrypt(
            key,
            props,
            zlib.compress(
                open("tests/res/msgstore.db", 'rb').read(),
                level=1,
            )
        )
        new_check = sha512(data).digest()
        with open("tests/res/msgstore-new.db.crypt14", 'wb') as f:
            f.write(data)
        with open("tests/res/msgstore-noexpiry.db.crypt14", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert new_check == orig_check
        os.remove("tests/res/msgstore-new.db.crypt14")

    def test_encryption12(self):
        key = KeyFactory.new("tests/res/key")
        props = Props(wa_version="2.22.5.13", jid="67", features=None)
        db = Database12(key=key, iv=bytes.fromhex("F4E9A6DC0B6F0D8986AF6C7180F02356"))
        data = db.encrypt(
            key,
            props,
            zlib.compress(
                open("tests/res/msgstore.db", 'rb').read(),
                level=1,
            )
        )
        new_check = sha512(data).digest()
        with open("tests/res/msgstore-new.db.crypt12", 'wb') as f:
            f.write(data)
        with open("tests/res/msgstore.db.crypt12", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert new_check == orig_check
        os.remove("tests/res/msgstore-new.db.crypt12")



#!/usr/bin/env python
"""
This script decrypts WhatsApp's DB files encrypted with Crypt12, Crypt14 or Crypt15.
"""

from __future__ import annotations

from wa_crypt_tools.lib.logformat import CustomFormatter
from wa_crypt_tools.lib.key.keyfactory import KeyFactory
from wa_crypt_tools.lib.db.dbfactory import DatabaseFactory
from wa_crypt_tools.lib.utils import test_decompression

# AES import party!
# pycryptodome and PyCryptodomex's implementations of AES are the same,
# so we try to import one of these twos.
try:
    # pycryptodomex
    from Cryptodome.Cipher import AES
except ModuleNotFoundError:
    try:
        # pycryptodome
        # noinspection PyUnresolvedReferences
        from Crypto.Cipher import AES

        if not hasattr(AES, 'MODE_GCM'):
            # pycrypto
            raise ModuleNotFoundError("You installed pycrypto and not pycryptodome(x). "
                                      "Pycrypto is old, deprecated and not supported. \n"
                                      "Run: python -m pip uninstall pycrypto\n"
                                      "And: python -m pip install pycryptodomex\n"
                                      "Or:  python -m pip install pycryptodome")
    except ModuleNotFoundError:
        # crypto (or nothing)
        raise ModuleNotFoundError("You need pycryptodome(x) to run these scripts!\n"
                                  "python -m pip install pycryptodome\n"
                                  "Or: python -m pip install pycryptodome\n"
                                  "You can also remove \"crypto\" if you have it installed\n"
                                  "python -m pip uninstall crypto")
# noinspection PyPackageRequirements
# This is from javaobj-py3

# noinspection PyPackageRequirements

import io
from re import findall
from sys import maxsize
from time import sleep
from datetime import date

import argparse
import zlib

__author__ = 'ElDavo'
__copyright__ = 'Copyright (C) 2023'
__license__ = 'GPLv3'
__status__ = 'Production'

import logging

log = logging.getLogger(__name__)


def parsecmdline() -> argparse.Namespace:
    """Sets up the argument parser"""
    parser = argparse.ArgumentParser(description='Decrypts WhatsApp backup files'
                                                 ' encrypted with crypt12, 14 or 15')
    parser.add_argument('keyfile', nargs='?', type=str, default="encrypted_backup.key",
                        help='The WhatsApp encrypted_backup key file or the hex encoded key. '
                             'Default: encrypted_backup.key')
    parser.add_argument('encrypted', nargs='?', type=argparse.FileType('rb'), default="msgstore.db.crypt15",
                        help='The encrypted crypt12, 14 or 15 file. Default: msgstore.db.crypt15')
    parser.add_argument('decrypted', nargs='?', type=argparse.FileType('wb'), default="msgstore.db",
                        help='The decrypted output file. Default: msgstore.db')
    parser.add_argument('-nm', '--no-mem', action='store_true',
                        help='Does not load files in RAM, stresses the disk more. '
                             'Default: load files into RAM')
    parser.add_argument('-bs', '--buffer-size', type=int, help='How many bytes of data to process at a time. '
                                                               'Implies -nm. Default: {}'.format(
        io.DEFAULT_BUFFER_SIZE))
    parser.add_argument('-nd', '--no-decompress', action='store_true',
                        help='Does not decompress the decrypted data. '
                             'Default: decompresses the decrypted data')
    parser.add_argument('-v', '--verbose', action='store_true', help='Prints all offsets and messages')
    parser.add_argument('-f', '--force', action='store_true', help='Does nothing, but it is here for compatibility')

    return parser.parse_args()


def chunked_decrypt(file_hash, cipher, encrypted, decrypted, buffer_size: int = 0, no_decompress: bool = False):
    """Does the actual decryption."""

    z_obj = zlib.decompressobj()

    if cipher is None:
        log.fatal("Could not create a decryption cipher")

    try:

        if buffer_size < 17:
            log.info("Invalid buffer size, will use default of {}".format(io.DEFAULT_BUFFER_SIZE))
            buffer_size = io.DEFAULT_BUFFER_SIZE

            # Does the thing above but only with DEFAULT_BUFFER_SIZE bytes at a time.
            # Less RAM used, more I/O used

            is_zip = True

            chunk = encrypted.read(buffer_size)

            log.debug("Reading and decrypting...")

            while next_chunk := encrypted.read(buffer_size):

                # We will need to manage two chunks at a time, because we might have
                # the checksum in both the last chunk and the chunk before that.
                # This makes the logic more complicated, but it's the only way to.

                checksum = None

                try:
                    next_chunk = encrypted.read(buffer_size)
                except MemoryError:
                    log.fatal("Out of RAM, please use a smaller buffer size.")

                if len(next_chunk) <= 36:
                    # Last bytes read. Three cases:
                    # 1. The checksum is entirely in the last chunk
                    if len(next_chunk) == 36:
                        checksum = next_chunk
                    # 2. The checksum is entirely in the chunk before the last
                    elif len(next_chunk) == 0:
                        checksum = chunk[-36:]
                        chunk = chunk[:-36]
                    # 3. The checksum is split between the last two chunks
                    else:
                        checksum = chunk[-(36 - len(next_chunk)):] + next_chunk
                        chunk = chunk[:-(36 - len(next_chunk))]

                file_hash.update(chunk)

                decrypted_chunk = cipher.decrypt(chunk)
                if is_zip:
                    try:
                        if no_decompress:
                            decrypted.write(decrypted_chunk)
                        else:
                            decrypted.write(z_obj.decompress(decrypted_chunk))
                    except zlib.error:
                        if test_decompression(decrypted_chunk):
                            log.info("Decrypted data is a ZIP file that I will not decompress automatically.")
                        else:
                            log.error("I can't recognize decrypted data. Decryption not successful.\n    "
                                    "The key probably does not match with the encrypted file.")
                        is_zip = False
                        decrypted.write(decrypted_chunk)
                else:
                    decrypted.write(decrypted_chunk)

                # The presence of the checksum tells us it's the last chunk
                if checksum is not None:
                    is_multifile_backup = False

                    crypt12_footer = str(checksum[-4:])
                    jid = findall(r"(?:-|\d)(?:-|\d)(\d\d)", crypt12_footer)
                    if len(jid) == 1:
                        # Confirmed to be crypt12
                        checksum = checksum[:-4]
                        log.debug("Your phone number ends with {}".format(jid[0]))
                    else:
                        # Shift everything forward by 4 bytes
                        chunk = checksum[:4]
                        file_hash.update(chunk)
                        decrypted_chunk = cipher.decrypt(chunk)
                        if is_zip:
                            try:
                                if no_decompress:
                                    decrypted.write(decrypted_chunk)
                                else:
                                    decrypted.write(z_obj.decompress(decrypted_chunk))
                            except zlib.error:
                                log.error("Backup is corrupted.")
                                decrypted.write(decrypted_chunk)
                        else:
                            decrypted.write(decrypted_chunk)
                        checksum = checksum[4:]

                    file_hash.update(checksum[:16])
                    if file_hash.digest() != checksum[16:]:
                        is_multifile_backup = True
                    else:
                        log.debug("Checksum OK ({})!".format(file_hash.hexdigest()))
                    try:
                        if is_multifile_backup:
                            decrypted.write(cipher.decrypt(checksum[:16]))
                            cipher.verify(checksum[16:])
                        else:
                            cipher.verify(checksum[:16])
                    except ValueError as e:
                        log.error("Authentication tag mismatch: {}."
                                "\n    This probably means your backup is corrupted.".format(e))
                    break

                chunk = next_chunk

            if is_zip and not z_obj.eof:
                log.error("The encrypted database file is truncated (damaged).")

        decrypted.flush()

    except OSError as e:
        log.fatal("I/O error: {}".format(e))

    finally:
        decrypted.close()
        encrypted.close()


def main():
    args = parsecmdline()

    # set wa_crypt_tools l to debug
    log.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch.setFormatter(CustomFormatter())
    log.addHandler(ch)
    # also add to "wa_crypt_tools.lib" logger
    logging.getLogger("wa_crypt_tools.lib").addHandler(ch)
    logging.getLogger("wa_crypt_tools.lib").setLevel(logging.DEBUG if args.verbose else logging.INFO)
    if args.buffer_size is not None:
        if not 1 < args.buffer_size < maxsize:
            log.fatal("Invalid buffer size")
    # Get the decryption key from the key file or the hex encoded string.
    key = KeyFactory.new(args.keyfile)
    log.debug(str(key))

    db = DatabaseFactory.from_file(args.encrypted)
    cipher = AES.new(key.get(), AES.MODE_GCM, db.get_iv())

    if args.buffer_size is not None:
        chunked_decrypt(db.file_hash, cipher, args.encrypted, args.decrypted, args.buffer_size, args.no_decompress)
    elif args.no_mem:
        chunked_decrypt(db.file_hash, cipher, args.encrypted, args.decrypted, io.DEFAULT_BUFFER_SIZE,
                        args.no_decompress)
    else:
        output_decrypted: bytearray = db.decrypt(key, args.encrypted.read())
        try:

            z_obj = zlib.decompressobj()
            if args.no_decompress:
                output_file = output_decrypted
            else:
                output_file = z_obj.decompress(output_decrypted)
                if not z_obj.eof:
                    log.error("The encrypted database file is truncated (damaged).")
        except zlib.error:
            output_file = output_decrypted
            if test_decompression(output_file[:io.DEFAULT_BUFFER_SIZE]):
                log.info("Decrypted data is a ZIP file that I will not decompress automatically.")
            else:
                log.error("I can't recognize decrypted data. Decryption not successful.\n    "
                        "The key probably does not match with the encrypted file.\n    "
                        "Or the backup is simply empty. (check with --force)")
        args.decrypted.write(output_file)

    if date.today().day == 1 and date.today().month == 4:
        log.info("Done. Uploading messages to the developer's server...")
        sleep(0.5)
        log.info("Uploaded. The developer will now read and publish your messages!")
    else:
        log.info("Done")


if __name__ == "__main__":
    main()



"""
This script tries to find protobuf messages in a file.
Props to protobuf_inspector for the parser.
"""

from __future__ import annotations

from protobuf_inspector.types import StandardParser
from io import BytesIO
from os.path import getsize

import argparse


def parsecmdline() -> argparse.Namespace:
    """Sets up the argument parser"""
    parser = argparse.ArgumentParser(description='Find protocol buffers in a file.')
    parser.add_argument('file_name', type=str, help='A file that you believe contains protobuf messages.')
    parser.add_argument('-k', '--keep-going', action='store_true', help='Don\'t stop after the first result.')
    parser.add_argument('-r', '--range', type=int, default=512, help='The number of bytes to search. Ignored if -w is '
                                                                     'set.')
    parser.add_argument('-w', '--whole-file', action='store_true', help='Search the whole file. Not advised for large '
                                                                        'files')
    return parser.parse_args()


def load_file(file_name: str, byte_range=0, reverse=False) -> bytes:
    """ Loads a file and returns it as a byte array.
    If byte_range is set, it will return the first byte_range bytes of the file.
    If reverse is set, it will return the last byte_range bytes of the file.
    """
    try:

        size = getsize(file_name)

        if byte_range > size / 2:
            raise ValueError("Range provided is bigger than half of file. Use -w or lower the range.")

        with open(file_name, 'rb') as f:

            if byte_range < 1:
                if reverse:
                    raise ValueError("Cannot read the whole file from end")
                return f.read()

            if reverse:
                f.seek(size - byte_range)
                return f.read(byte_range)

            return f.read(byte_range)

    except IOError:
        print("File not found or other IO error")
        exit(1)


def get_truncated_stream(content: bytes, start: int, end: int) -> BytesIO:
    """ Returns a BytesIO object with the content truncated to the given range. """
    return BytesIO(content[start:end])


def main():
    args = parsecmdline()

    if args.whole_file:

        whole_file = load_file(args.file_name)
        search(whole_file, args.keep_going)

    else:

        # We first try the first "range" bytes.
        whole_file = load_file(args.file_name, args.range)
        search(whole_file, args.keep_going)

        print("Now searching at the end of the file")
        # Then we try the last "range" bytes.
        size = getsize(args.file_name)
        whole_file = load_file(args.file_name, args.range, reverse=True)
        search(whole_file, args.keep_going, size - len(whole_file))

    if args.keep_going:
        print("Finished")
    else:
        print("Nothing found")
        exit(1)


def protoparse(stream):
    return StandardParser().parse_message(stream, "message")


class Message:
    last_good_i = 0
    last_good_j = 0
    output = ""


def search(whole_file: bytes, keep_going: bool, offset=0):
    """ Searches for protobuf messages in the given byte array. """

    for i in range(len(whole_file)):
        message = Message()

        for j in range(i + 1, len(whole_file)):
            candidate = get_truncated_stream(whole_file, i, j)
            try:
                message.output = protoparse(candidate)
                message.last_good_j = j
                message.last_good_i = i

            except Exception:
                pass
        if message.last_good_j and message.last_good_i == i:
            print("\n Message from byte {} to {}".format(message.last_good_i + offset, message.last_good_j + offset))
            print(message.output)
            if not keep_going:
                exit(0)


if __name__ == "__main__":
    main()



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 27, 3, '', 'C14_cipher.proto')
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10C14_cipher.proto"m\n\nC14_cipher\x12\x16\n\x0ecipher_version\x18\x01 \x01(\x0c\x12\x13\n\x0bkey_version\x18\x02 \x01(\x0c\x12\x13\n\x0bserver_salt\x18\x03 \x01(\x0c\x12\x11\n\tgoogle_id\x18\x04 \x01(\x0c\x12\n\n\x02IV\x18\x05 \x01(\x0cb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'C14_cipher_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_C14_CIPHER']._serialized_start = 20
    _globals['_C14_CIPHER']._serialized_end = 129


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 27, 3, '', 'C15_IV.proto')
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cC15_IV.proto"\x14\n\x06C15_IV\x12\n\n\x02IV\x18\x01 \x01(\x0cb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'C15_IV_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_C15_IV']._serialized_start = 16
    _globals['_C15_IV']._serialized_end = 36


import os
import zlib

from wa_crypt_tools.lib.key.key15 import Key15
from wa_crypt_tools.lib.key.key14 import Key14
from wa_crypt_tools.lib.key.keyfactory import KeyFactory
from wa_crypt_tools.lib.props import Props
from hashlib import sha512


class TestCreatekey:
    def test_createkey(self):
        key: Key15 = Key15(key=
        bytes.fromhex(
            '6730a595a1484d0c39c101dc0ac82ec5e401bb6f0e1b8ee2dc104a6b3687f017'
        ))
        keyb: bytes = key.dump()
        keyb_digest = sha512(keyb).digest()
        with open("tests/res/encrypted_backup.key", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert keyb_digest == orig_check

    def test_createkey14(self):
        key: Key14 = Key14(key=
        bytes.fromhex(
            '3a146d9bbd8b6311d962c71619c0c2cce3ce694ea4a0f3f600e271380e1226c6'
        ),
            serversalt=bytes.fromhex('cd788b1b4625f50d3fccdeac94e1ff638899733b77a224ff614918363901f044'),
            googleid=bytes.fromhex('92683e735c88727eef9486911f3ac6fa'),
            key_version=b'\x02',
            cipher_version=b'\x00\x01')
        keyb: bytes = key.dump()
        keyb_digest = sha512(keyb).digest()
        with open("tests/res/key", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert keyb_digest == orig_check



from pathlib import Path

import javaobj.v2 as javaobj

from wa_crypt_tools.lib.key.key14 import Key14
from wa_crypt_tools.lib.key.key15 import Key15

import logging

from wa_crypt_tools.lib.utils import javaintlist2bytes, hexstring2bytes

l = logging.getLogger(__name__)
class KeyFactory:
    @staticmethod
    def new(file: Path):
        """Tries to load the key from a file, or if it fails, from a hex string."""
        try:
            return KeyFactory.from_file(file)
        except OSError:
            try:
                return KeyFactory.from_hex(str(file))
            except ValueError:
                l.critical("The key file specified does not exist.\n    "
                           "If you tried to specify the key directly, note it should be "
                           "64 characters long and not {} characters long.".format(len(str(file))))

    @staticmethod
    def from_file(file: Path):
        keyfile: bytes = b''

        l.debug("Reading keyfile...")

        # Try to open the keyfile.
        try:
            key_file_stream = open(file, 'rb')
            try:
                # Deserialize the byte object written in the file
                jarr: javaobj.beans.JavaArray = javaobj.load(key_file_stream).data
                # Convert from a list of Int8 to a byte array
                keyfile: bytes = javaintlist2bytes(jarr)

            except (ValueError, RuntimeError) as e:
                l.critical("The keyfile is not a valid Java object: {}".format(e))

        except OSError:
            l.info("The keyfile could not be opened.")
            raise OSError

        # We guess the key type from its length
        if len(keyfile) == 131:
            return Key14(keyarray=keyfile)
        elif len(keyfile) == 32:
            return Key15(keyarray=keyfile)
        else:
            l.critical("Unrecognized key file format.")

    @staticmethod
    def from_hex(hexstring: str) -> Key15:
        if hexstring is None or len(hexstring) != 64:
            raise ValueError("The key is invalid or of the wrong length.")
        barr: bytes = hexstring2bytes(hexstring)
        if barr is None or len(barr) != 32:
            raise ValueError("The key is invalid or of the wrong length.")
        return Key15(keyarray=barr)


from __future__ import annotations

from wa_crypt_tools.lib.constants import C
from wa_crypt_tools.proto import backup_expiry_pb2 as backup_expiry

class Props:
    def __init__(self, *, v_features=None, wa_version: str = C.DEFAULT_APP_VERSION, jid: str = C.DEFAULT_JID_SUFFIX,
                 features: list[int] | None = C.DEFAULT_FEATURE_LIST, max_feature: int = C.DEFAULT_MAX_FEATURE,
                 backup_version: int = C.DEFAULT_BACKUP_VERSION):
        if v_features is not None:
            self.props = v_features
            return
        self.props = backup_expiry.BackupExpiry()
        self.props.app_version = wa_version
        self.props.jidSuffix = jid
        self.max_feature = max_feature
        if features is None or len(features) == 0:
            return
        self.props.backup_version = backup_version
        for f in range(5, max_feature + 1):
            try:
                self.disable_feature(f)
            except AttributeError:
                pass
        for f in features:
            self.enable_feature(f)

    def enable_feature(self, feature: int):
        feature_name = "f_" + str(feature)
        setattr(self.props, feature_name, True)

    def disable_feature(self, feature: int):
        feature_name = "f_" + str(feature)
        setattr(self.props, feature_name, False)

    def get_feature(self, feature: int) -> bool:
        feature_name = "f_" + str(feature)
        return getattr(self.props, feature_name)

    def get_features(self) -> list[int]:
        features = []
        for i in range(5, self.max_feature + 1):
            try:
                if self.get_feature(i):
                    features.append(i)
            except AttributeError:
                pass
        return features

    def get_wa_version(self) -> str:
        return self.props.version

    def get_jid(self) -> str:
        return self.props.jidSuffix

    def get_proto(self):
        return self.props

    def __str__(self):
        return str(self.props)


from __future__ import annotations

import abc

class Key(abc.ABC):
    @abc.abstractmethod
    def __init__(self, keyarray: bytes = None):
        pass

    @abc.abstractmethod
    def __str__(self) -> str:
        pass

    @abc.abstractmethod
    def get(self) -> bytes:
        pass

    @abc.abstractmethod
    def dump(self) -> bytes:
        pass



import logging

from google.protobuf.message import DecodeError

from wa_crypt_tools.lib.constants import C
from wa_crypt_tools.lib.db.db12 import Database12
from wa_crypt_tools.lib.db.db14 import Database14
from wa_crypt_tools.lib.db.db15 import Database15
from wa_crypt_tools.lib.props import Props
from wa_crypt_tools.lib.utils import header_info

log = logging.getLogger(__name__)

from hashlib import md5
from re import findall


class DatabaseFactory:
    @staticmethod
    def from_file(encrypted):
        try:
            from wa_crypt_tools.proto import backup_prefix_pb2 as prefix
            from wa_crypt_tools.proto import key_type_pb2 as key_type
        except ImportError as e:
            log.error("Could not import the proto classes: {}".format(e))
            if str(e).startswith("cannot import name 'builder' from 'google.protobuf.internal'"):
                log.error("You need to upgrade the protobuf library to at least 3.20.0.\n"
                          "    python -m pip install --upgrade protobuf")
            elif str(e).startswith("no module named"):
                log.error("Please download them and put them in the \"proto\" sub folder.")
            raise e
        except AttributeError as e:
            log.error("Could not import the proto classes: {}\n    ".format(e) +
                      "Your protobuf library is probably too old.\n    "
                      "Please upgrade to at least version 3.20.0 , by running:\n    "
                      "python -m pip install --upgrade protobuf")
            raise e

        header = prefix.BackupPrefix()

        log.debug("Parsing database header...")

        try:
            file_hash = md5()
            # The first byte is the size of the upcoming protobuf message
            protobuf_size = encrypted.read(1)
            file_hash.update(protobuf_size)
            protobuf_size = int.from_bytes(protobuf_size, byteorder='big')

            # A 0x01 as a second byte indicates the presence of the feature table in the protobuf.
            # It is optional and present only in msgstore database, although
            # I found some old msgstore backups without it, so it is optional.
            msgstore_features_flag = encrypted.peek(1)[0]
            if msgstore_features_flag != 1:
                msgstore_features_flag = 0
            else:
                file_hash.update(encrypted.read(1))
            if not msgstore_features_flag:
                log.debug("No feature table found (not a msgstore DB or very old)")

            try:

                protobuf_raw = encrypted.read(protobuf_size)
                file_hash.update(protobuf_raw)

                if header.ParseFromString(protobuf_raw) != protobuf_size:
                    log.error("Protobuf message not fully read. Please report a bug.")
                else:

                    # Checking and printing WA version and phone number
                    version = findall(r"\d(?:\.\d{1,3}){3}", header.info.app_version)
                    if len(version) != 1:
                        log.error('WhatsApp version not found')
                    else:
                        log.debug("WhatsApp version: {}".format(version[0]))
                    if len(header.info.jidSuffix) != 2:
                        log.error("The phone number end is not 2 characters long")
                    log.debug("Your phone number ends with {}".format(header.info.jidSuffix))

                    if len(header.c15_iv.IV) != 0:
                        # DB Header is crypt15
                        # if type(key) is not Key15:
                        #    l.error("You are using a crypt14 key file with a crypt15 backup.")
                        if len(header.c15_iv.IV) != 16:
                            log.error("IV is not 16 bytes long but is {} bytes long".format(len(header.c15_iv.IV)))
                        iv = header.c15_iv.IV

                    elif len(header.c14_cipher.IV) != 0:

                        # DB Header is crypt14
                        # if type(key) is not Key14:
                        #    l.fatal("You are using a crypt15 key file with a crypt14 backup.")

                        # if key.cipher_version != p.c14_cipher.version.cipher_version:
                        #    l.error("Cipher version mismatch: {} != {}"
                        #    .format(key.cipher_version, p.c14_cipher.cipher_version))

                        # Fix bytes to string encoding key.key_version = (key.key_version[0] + 48).to_bytes(1,
                        # byteorder='big') if key.key_version != p.c14_cipher.key_version: if key.key_version >
                        # p.c14_cipher.key_version: l.error("Key version mismatch: {} != {} .\n    " .format(
                        # key.key_version, p.c14_cipher.key_version) + "Your backup is too old for this key file.\n
                        # " + "Please try using a newer backup.") elif key.key_version < p.c14_cipher.key_version:
                        # l.error("Key version mismatch: {} != {} .\n    " .format(key.key_version,
                        # p.c14_cipher.key_version) + "Your backup is too new for this key file.\n    " + "Please try
                        # using an older backup, or getting the new key.") else: l.error("Key version mismatch: {} !=
                        # {} (?)" .format(key.key_version, p.c14_cipher.key_version)) if key.get_serversalt() !=
                        # p.c14_cipher.server_salt: l.error("Server salt mismatch: {} != {}".format(
                        # key.get_serversalt(), p.c14_cipher.server_salt)) if key.get_googleid() !=
                        # p.c14_cipher.google_id: l.error("Google ID mismatch: {} != {}".format(key.get_googleid(),
                        # p.c14_cipher.google_id))
                        if len(header.c14_cipher.IV) != 16:
                            log.error(
                                "IV is not 16 bytes long but is {} bytes long".format(len(header.c14_cipher.IV)))
                        iv = header.c14_cipher.IV

                    else:
                        log.error("Could not parse the IV from the protobuf message. Please report a bug.")
                        raise DecodeError

                    # We are done here
                    log.debug(header_info(header))

                    props = Props(v_features=header.info)
                    if header.c15_iv.IV:
                        db = Database15(iv=iv, props=props)
                        db.file_hash = file_hash
                        return db
                    elif header.c14_cipher.IV:
                        db = Database14(iv=iv, props=props)
                        db.file_hash = file_hash
                        return db
                    else:
                        log.error("Could not parse the IV from the protobuf message. Please report a bug.")
                        raise DecodeError

            except DecodeError:

                # try again as a crypt12
                log.debug("Could not parse the protobuf message as a crypt14/15. Trying as a crypt12...")
                try:
                    encrypted.seek(0)
                except OSError as e:
                    log.fatal("Could not reset the file pointer: {}".format(e))
                    raise e
                return Database12(encrypted=encrypted)

        except OSError as e:
            log.fatal("Reading database header failed: {}".format(e))



from hashlib import md5
from os import urandom
from pathlib import Path
from re import findall

from Cryptodome.Cipher import AES

import logging

from wa_crypt_tools.lib.constants import C
from wa_crypt_tools.lib.db.db import Database
from wa_crypt_tools.lib.key.key14 import Key14
from wa_crypt_tools.lib.props import Props

log = logging.getLogger(__name__)


class Database12(Database):
    """
    Implementation of a crypt12 database.
    """

    def __init__(self, key: Key14 = None, encrypted=None,
                 cipher_version: bytes = None, key_version: bytes = None, serversalt: bytes = None,
                 googleid: bytes = None, iv: bytes = None):
        """Checks if the file is a Crypt12 file.
        Returns the cipher if it is, None otherwise."""

        """
        The crypt12 file format is similar to the crypt14 file format.
        It is a "raw" header, which means it's not a protobuf message,
        nor a serialized java object.
        Structure:
        Cipher version (2 bytes)
        Key version (1 byte)
        Server salt (32 bytes)
        Google ID (16 bytes)
        IV (16 bytes)
        ( so we finally understood why the IV is at offset 51 ... )
        """
        self.file_hash = md5()
        if encrypted and key:
            self.cipher_version = encrypted.read(2)
            if self.cipher_version != key.get_cipher_version():
                log.error("Cipher version mismatch: {} != {}".format(self.cipher_version, key.get_cipher_version()))
                raise ValueError
            self.file_hash.update(self.cipher_version)

            self.key_version = encrypted.read(1)
            if self.key_version != key.get_key_version():
                log.error("Key version mismatch: {} != {}".format(self.key_version, key.get_key_version()))
                raise ValueError
            self.file_hash.update(self.key_version)

            self.serversalt = encrypted.read(32)
            if self.serversalt != key.get_serversalt():
                log.error("Server salt mismatch: {} != {}".format(self.serversalt, key.get_serversalt()))
                raise ValueError
            self.file_hash.update(self.serversalt)

            self.googleid = encrypted.read(16)
            if self.googleid != key.get_googleid():
                log.error("Google ID mismatch: {} != {}".format(self.googleid, key.get_googleid()))
            self.file_hash.update(self.googleid)

            self.iv = encrypted.read(16)
            self.file_hash.update(self.iv)
        elif encrypted:
            self.cipher_version = encrypted.read(2)
            # if test_bytes != key.get_cipher_version():
            #    quit_12()
            self.file_hash.update(self.cipher_version)

            self.key_version = encrypted.read(1)
            # if test_bytes != key.get_key_version():
            #    quit_12()
            self.file_hash.update(self.key_version)

            self.serversalt = encrypted.read(32)
            # if test_bytes != key.get_serversalt():
            #    quit_12()
            self.file_hash.update(self.serversalt)

            self.googleid = encrypted.read(16)
            # if test_bytes != key.get_googleid():
            #    quit_12()
            self.file_hash.update(self.googleid)

            self.iv = encrypted.read(16)
            self.file_hash.update(self.iv)
        elif key:
            self.cipher_version = key.get_cipher_version()
            self.file_hash.update(self.cipher_version)
            self.key_version = key.get_key_version()
            self.file_hash.update(self.key_version)
            self.serversalt = key.get_serversalt()
            self.file_hash.update(self.serversalt)
            self.googleid = key.get_googleid()
            self.file_hash.update(self.googleid)
            if iv:
                self.iv = iv
            else:
                self.iv = urandom(16)
            self.file_hash.update(self.iv)
        else:
            if cipher_version:
                if cipher_version == C.SUPPORTED_CIPHER_VERSION:
                    self.cipher_version = cipher_version
                    self.file_hash.update(self.cipher_version)
                else:
                    log.error("Unsupported cipher version provided!")
                    raise ValueError
            else:
                self.cipher_version = C.SUPPORTED_CIPHER_VERSION
                self.file_hash.update(self.cipher_version)

            if key_version:
                if key_version in C.SUPPORTED_KEY_VERSIONS:
                    self.key_version = key_version
                    self.file_hash.update(self.key_version)
                else:
                    log.error("Unsupported key version provided!")
            else:
                self.key_version = C.SUPPORTED_KEY_VERSIONS[-1]
                self.file_hash.update(self.key_version)

            if serversalt:
                self.serversalt = serversalt
            else:
                self.serversalt = urandom(32)
            self.file_hash.update(self.serversalt)

            if googleid:
                self.googleid = googleid
            else:
                self.googleid = urandom(16)
            self.file_hash.update(self.googleid)

            if iv:
                self.iv = iv
            else:
                self.iv = urandom(16)
            self.file_hash.update(self.iv)

    def __str__(self):
        return f"""cipher_version: {self.cipher_version}
                    key_version: {self.key_version}
                    serversalt: {self.serversalt}
                    googleid: {self.googleid}
                    iv: {self.iv}"""

    def decrypt(self, key: Key14, encrypted: bytes) -> bytes:
        """Decrypts the database using the provided key"""
        userjid = encrypted[-4:]
        # check the userjid
        crypt12_footer = str(userjid)
        jid = findall(r"(?:-|\d)(?:-|\d)(\d\d)", crypt12_footer)
        if len(jid) != 1:
            log.error("The phone number end is not 2 characters long")
        else:
            log.debug("Your phone number ends with {}".format(jid[0]))
        checksum = encrypted[-20:-4]
        authentication_tag = encrypted[-36:-20]
        encrypted_data = encrypted[:-36]
        is_multifile_backup = False

        self.file_hash.update(encrypted_data)
        self.file_hash.update(authentication_tag)

        if self.file_hash.digest() != checksum:
            # We are probably in a multifile backup, which does not have a checksum.
            # TODO do crypt12 multifiles actually exist?
            is_multifile_backup = True
        else:
            log.debug("Checksum OK ({}). Decrypting...".format(self.file_hash.hexdigest()))

        cipher = AES.new(key.get(), AES.MODE_GCM, self.iv)
        try:
            output_decrypted: bytes = cipher.decrypt(encrypted_data)
        except ValueError as e:
            log.fatal("Decryption failed: {}."
                      "\n    This probably means your backup is corrupted.".format(e))
            raise e

        # Verify the authentication tag
        try:
            if is_multifile_backup:
                # In multifile backups, there is no checksum.
                # This means, the last 16 bytes of the files are not the checksum,
                # despite being called "checksum", but are the authentication tag.
                # Same way, "authentication tag" is not the tag, but the last
                # 16 bytes of the encrypted file.
                output_decrypted += cipher.decrypt(authentication_tag)
                cipher.verify(checksum)
            else:
                cipher.verify(authentication_tag)
        except ValueError as e:
            log.error("Authentication tag mismatch: {}."
                      "\n    This probably means your backup is corrupted.".format(e))

        return output_decrypted

    def encrypt(self, key: Key14, props: Props, decrypted: bytes) -> bytes:
        file_hash = md5()
        out = b""
        out += self.cipher_version
        out += self.key_version
        out += self.serversalt
        out += self.googleid
        out += self.iv
        cipher = AES.new(key.get(), AES.MODE_GCM, self.iv)
        encrypted = cipher.encrypt(decrypted)
        out += encrypted
        out += cipher.digest()
        file_hash.update(out)
        out += file_hash.digest()
        jid = props.get_jid()
        if len(jid) != 2:
            log.error("The phone number end is not 2 characters long")
        out += "--{}".format(jid).encode()
        return out

    def get_iv(self) -> bytes:
        return self.iv



import abc
import logging

from wa_crypt_tools.lib.key.key import Key
from wa_crypt_tools.lib.props import Props

log = logging.getLogger(__name__)


class Database(abc.ABC):
    """
    An abstract class that represents a database.
    """
    iv: bytes

    @abc.abstractmethod
    def __str__(self):
        pass

    @abc.abstractmethod
    def decrypt(self, key: Key, encrypted: bytes) -> bytes:
        pass

    @abc.abstractmethod
    def encrypt(self, key: Key, props: Props, decrypted: bytes) -> bytes:
        pass

    @abc.abstractmethod
    def get_iv(self) -> bytes:
        return self.iv



#!/usr/bin/env python
"""
This script transforms a password_data.key file into a hashcat hash.
"""
from __future__ import annotations

import argparse

# noinspection PyPackageRequirements
# This is from javaobj-py3
import javaobj.v2 as javaobj

from base64 import b64encode


# password_data.key file format:
# The password_data.key file is a serialized Java object, composed by:
# 1) An int (4 bytes), which should be the version of the format.
# The only known value for now is 1, so we only support this version.
# 2) The encoded password (32 bytes), encoded with PBKDF2-HMAC-SHA512.
# PBKDF2 needs a salt and a permutation number, which are written after:
# 3) The salt (32 bytes), which is a random byte array.
# 4) The permutation number (4 bytes), which is an int. For now seems to be fixed at 100000.
# The script parses the permutation number for the file, so if it changes, no problem.

class Log:
    """Simpler logger class. Supports 2 verbosity levels."""

    @staticmethod
    def i(msg: str):
        """Always prints message."""
        print('[I] {}'.format(msg))

    @staticmethod
    def f(msg: str):
        """Always prints message and exit."""
        print('[F] {}'.format(msg))
        exit(1)


def parsecmdline() -> argparse.Namespace:
    """Sets up the argument parser"""
    parser = argparse.ArgumentParser(description='Gives a hashcat representation of the password data key')
    parser.add_argument('passworddatakeyfile', nargs='?', type=argparse.FileType('rb'), default="password_data.key",
                        help='The WhatsApp password data keyfile. Default: password_data.key')
    return parser.parse_args()


def barrtoint(barr: javaobj.beans.BlockData) -> int:
    """Converts a javaobj BlockData to an int"""
    return int.from_bytes(barr.data, byteorder='big', signed=False)


def javaintlist2bytes(barr: javaobj.beans.JavaArray) -> bytes:
    """Converts a javaobj bytearray which somehow became a list of signed integers back to a Python byte array"""
    out: bytes = b''
    for i in barr.data:
        out += i.to_bytes(1, byteorder='big', signed=True)
    return out


def read_password_data_key(passworddatakeyfilestream) -> str:
    # Assign variables to suppress warnings
    deserialized: list = list()

    try:
        deserialized: list = javaobj.load(passworddatakeyfilestream)
    except OSError as e:
        Log.f("Couldn't read keyfile: {}".format(e))
    except (ValueError, RuntimeError) as e:
        Log.f("The keyfile is not a valid Java object: {}".format(e))

    if len(deserialized) != 4:
        Log.f("The keyfile has more fields than expected.")

    version: int = barrtoint(deserialized[0])
    if version != 1:
        Log.f("Unexpected key version: {}".format(version))

    encoded = javaintlist2bytes(deserialized[1])
    if len(encoded) != 64:
        Log.f("The encoded password has the wrong length")

    salt = javaintlist2bytes(deserialized[2])
    if len(salt) != 64:
        Log.f("The salt has the wrong length")

    permutations: int = barrtoint(deserialized[3])
    if permutations != 100000:
        Log.i("Unexpected permutation number: {}".format(permutations))

    return "sha512:{}:{}:{}".format(
        permutations,
        b64encode(salt).decode('ascii'),
        b64encode(encoded).decode('ascii')
    )


def main():
    args = parsecmdline()
    Log.i("Remember: hashcat mode is 12100 (PBKDF2-HMAC-SHA512)")
    pwd_hash = read_password_data_key(args.passworddatakeyfile)
    print(pwd_hash)


if __name__ == "__main__":
    main()



import base64
import hmac
import json
import math
import zlib
from hashlib import sha256

from Cryptodome.Cipher import AES
from javaobj import JavaByteArray
from javaobj.v2.beans import JavaArray, JavaClassDesc, ClassDescType

import logging

from wa_crypt_tools.lib.constants import C

# FIXME a "utils" file shouldn't have its own logger
l = logging.getLogger(__name__)


def test_decompression(test_data: bytes) -> bool:
    """Returns true if the SQLite header is valid.
    It is assumed that the data are valid.
    (If it is valid, it also means the decryption and decompression were successful.)"""

    # If we get a ZIP file header, return true
    if test_data[:4] == C.ZIP_HEADER:
        return True

    try:
        zlib_obj = zlib.decompressobj().decompress(test_data)
        # These two errors should never happen
        if len(zlib_obj) < 16:
            l.error("Test decompression: chunk too small")
            return False
        # Decoding can fail if first two bytes are a bad UTF-8 char
        if zlib_obj[:15].decode('ascii') != 'SQLite format 3':
            l.error("Test decompression: Decryption and decompression ok but not a valid SQLite database")
            return False
        else:
            return True
    except (zlib.error, UnicodeDecodeError):
        return False


def create_jba(out: bytes) -> JavaByteArray:
    """Creates a JavaByteArray object from a bytes array"""
    # Create the classdesc
    cd = JavaClassDesc(ClassDescType.NORMALCLASS)
    cd.name = "[B"
    cd.superclass = None
    cd.serial_version_uid = -5984413125824719648
    cd.desc_flags = 2

    return JavaByteArray(out, classdesc=cd)


def hexstring2bytes(string: str) -> bytes:
    """Converts a hex string into a bytes array"""
    if len(string) != 64:
        l.critical("The key file specified does not exist.\n    "
                   "If you tried to specify the key directly, note it should be "
                   "64 characters long and not {} characters long.".format(len(string)))

    barr = None
    try:
        barr = bytes.fromhex(string)
    except ValueError as e:
        l.critical("Couldn't convert the hex string.\n    "
                   "Exception: {}".format(e))
    if len(barr) != 32:
        l.error("The key is not 32 bytes long but {} bytes long.".format(len(barr)))
    return barr


def javaintlist2bytes(barr: JavaArray) -> bytes:
    """Converts a javaobj bytearray which somehow became a list of signed integers back to a Python byte array"""
    out: bytes = b''
    for i in barr:
        out += i.to_bytes(1, byteorder='big', signed=True)
    return out


def encryptionloop(*, first_iteration_data: bytes, privateseed: bytes = b'\x00' * 32, message: bytes,
                   output_bytes: int):
    # The private key and the seed are used to create the HMAC key
    privatekey = hmac.new(privateseed, msg=first_iteration_data, digestmod=sha256).digest()

    data = b''
    output = b''
    permutations = int(math.ceil(float(output_bytes) / float(32)))
    i = 1
    while i < permutations + 1:
        hasher = hmac.new(privatekey, msg=data, digestmod=sha256)
        if message is not None:
            hasher.update(message)
        hasher.update(i.to_bytes(1, byteorder='big'))
        data = hasher.digest()
        bytestowrite = min(output_bytes, len(data))
        output += data[:bytestowrite]
        i += 1
    return output


def mcrypt1_metadata_decrypt(*, key, encoded: str):
    """
    Decrypts the metadata of a mcrypt1 file.
    :param key: The key used to decrypt the metadata
    :param encoded: The metadata downloaded from Google Drive in base64
    :return: The decrypted JSON
    """
    # Base64 decoding
    encoded = base64.b64decode(encoded)
    # PKCS5Padding is not natively supported
    unpad = lambda s: s[:-ord(s[len(s) - 1:])]
    iv_size = encoded[0]
    if iv_size != 16:
        raise Exception("IV Size is not 16")

    iv = encoded[1:17]
    mac_size = encoded[17]
    if mac_size != 32:
        raise Exception("MAC Size is not 32")

    mac = encoded[18:50]
    encrypted_metadata = encoded[50:]
    # Authentication part
    hmac_auth = hmac.new(key.get_metadata_authentication(), digestmod='sha256')
    hmac_auth.update(iv)
    hmac_auth.update(encrypted_metadata)
    hmac_auth = hmac_auth.digest()
    if hmac_auth != mac:
        raise ValueError("MAC does not match")
    # Decryption part
    cipher = AES.new(key.get_metadata_encryption(), AES.MODE_CBC, iv)
    decrypted_metadata = cipher.decrypt(encrypted_metadata)
    decrypted_metadata = unpad(decrypted_metadata)
    # Load the JSON
    return json.loads(decrypted_metadata.decode('utf-8'))


def get_mcrypt1_name(*, key, name: str, md5: bytes) -> bytes:
    hmac_n = hmac.new(key.get_root(), digestmod='sha256')
    # Calculate SHA256 of the name
    digest = sha256()
    digest.update(name.encode('utf-8'))
    # Pour it into the HMAC
    hmac_n.update(digest.digest())
    # If md5 is a string, convert it to bytes
    if isinstance(md5, str):
        md5 = bytes.fromhex(md5)
    # Now pour the MD5 into the HMAC
    hmac_n.update(md5)
    media_hash = hmac_n.digest()
    return media_hash


def header_info(header):
    """
    shows all header, information including the feature vector
    FIXME
    """
    string: str = ""
    if header.c15_iv.IV:
        string += "Crypt15 info:\n"
        string += str("Header information in your crypt15 file:")
        string += str("IV: {}\n".format(header.c15_iv.IV.hex()))
    if header.c14_cipher.IV:
        string += str("Header information in your crypt14 file:\n")
        string += str("Cipher version: {}\n".format(header.c14_cipher.cipher_version.hex()))
        string += str("Key version: {}\n".format(header.c14_cipher.key_version.hex()))
        string += str("Server salt: {}\n".format(header.c14_cipher.server_salt.hex()))
        string += str("Google ID: {}\n".format(header.c14_cipher.google_id.hex()))
        string += str("IV: {}\n".format(header.c14_cipher.IV.hex()))
    string += str("Key type: {}\n".format(header.key_type))
    string += str("WhatsApp version: {}\n".format(header.info.app_version))
    #string += str("Device model: {}".format(header.info.device_model))
    string += str("The last two numbers of the user's Jid: {}\n".format(header.info.jidSuffix))
    string += str("Backup version: {}\n".format(header.info.backup_version))
    #string += str("Size of the backup file: {}".format(header.backup_export_file_size))
    features = [n for n in [*range(5, 38), 39] if getattr(header.info, "f_" + str(n)) == True]
    if len(features) > 0:
        string += str("Features: {}\n".format(features))
        string += str("Max feature number: {}\n".format(max(features)))
    else:
        string += str("No feature table found (not a msgstore DB or very old)\n")

    return string



import logging
from hashlib import md5
from os import urandom
from re import findall

from Cryptodome.Cipher import AES
from google.protobuf.message import DecodeError

from wa_crypt_tools.lib.props import Props

log = logging.getLogger(__name__)

from wa_crypt_tools.lib.db.db import Database
from wa_crypt_tools.lib.key.key15 import Key15


class Database15(Database):
    def __str__(self):
        return "Database15"
        # todo

    def __init__(self, *, key: Key15 = None, encrypted=None, iv: bytes = None
                 , props: Props = None):
        self.file_hash = md5()
        # just store it for now
        self.props = props
        if encrypted:
            try:
                from wa_crypt_tools.proto import backup_prefix_pb2 as prefix
                from wa_crypt_tools.proto import key_type_pb2 as key_type
            except ImportError as e:
                log.error("Could not import the proto classes: {}".format(e))
                if str(e).startswith("cannot import name 'builder' from 'google.protobuf.internal'"):
                    log.error("You need to upgrade the protobuf library to at least 3.20.0.\n"
                              "    python -m pip install --upgrade protobuf")
                elif str(e).startswith("no module named"):
                    log.error("Please download them and put them in the \"proto\" sub folder.")
                raise e
            except AttributeError as e:
                log.error("Could not import the proto classes: {}\n    ".format(e) +
                          "Your protobuf library is probably too old.\n    "
                          "Please upgrade to at least version 3.20.0 , by running:\n    "
                          "python -m pip install --upgrade protobuf")
                raise e

            self.header = prefix.BackupPrefix()

            log.debug("Parsing database header...")

            try:

                # The first byte is the size of the upcoming protobuf message
                protobuf_size = encrypted.read(1)
                self.file_hash.update(protobuf_size)
                protobuf_size = int.from_bytes(protobuf_size, byteorder='big')

                # A 0x01 as a second byte indicates the presence of the feature table in the protobuf.
                # It is optional and present only in msgstore database, although
                # I found some old msgstore backups without it, so it is optional.
                msgstore_features_flag = encrypted.peek(1)[0]
                if msgstore_features_flag != 1:
                    msgstore_features_flag = 0
                else:
                    self.file_hash.update(encrypted.read(1))
                if not msgstore_features_flag:
                    log.debug("No feature table found (not a msgstore DB or very old)")

                try:

                    protobuf_raw = encrypted.read(protobuf_size)
                    self.file_hash.update(protobuf_raw)

                    if self.header.ParseFromString(protobuf_raw) != protobuf_size:
                        log.error("Protobuf message not fully read. Please report a bug.")
                    else:

                        # Checking and printing WA version and phone number
                        version = findall(r"\d(?:\.\d{1,3}){3}", self.header.info.app_version)
                        if len(version) != 1:
                            log.error('WhatsApp version not found')
                        else:
                            log.debug("WhatsApp version: {}".format(version[0]))
                        if len(self.header.info.jidSuffix) != 2:
                            log.error("The phone number end is not 2 characters long")
                        log.debug("Your phone number ends with {}".format(self.header.info.jidSuffix))

                        if len(self.header.c15_iv.IV) != 0:
                            # DB Header is crypt15
                            # if type(key) is not Key15:
                            #    l.error("You are using a crypt14 key file with a crypt15 backup.")
                            if len(self.header.c15_iv.IV) != 16:
                                log.error(
                                    "IV is not 16 bytes long but is {} bytes long".format(len(self.header.c15_iv.IV)))
                            iv = self.header.c15_iv.IV

                        elif len(self.header.c14_cipher.IV) != 0:
                            raise ValueError("Crypt14 file in crypt15 constructor!")
                        else:
                            log.error("Could not parse the IV from the protobuf message. Please report a bug.")
                            raise ValueError

                except DecodeError as e:

                    log.error("Could not parse the protobuf message: {}".format(e))
                    raise e

            except OSError as e:
                log.fatal("Reading database header failed: {}".format(e))
                raise e
        else:
            if iv:
                if len(iv) != 16:
                    log.error("IV is not 16 bytes long but is {} bytes long".format(len(iv)))
                self.iv = iv
            else:
                self.iv = urandom(16)

    def decrypt(self, key: Key15, encrypted: bytes) -> bytes:
        """Decrypts the database using the provided key"""
        checksum = encrypted[-16:]
        authentication_tag = encrypted[-32:-16]
        encrypted_data = encrypted[:-32]
        is_multifile_backup = False

        self.file_hash.update(encrypted_data)
        self.file_hash.update(authentication_tag)

        if self.file_hash.digest() != checksum:
            # We are probably in a multifile backup, which does not have a checksum.
            is_multifile_backup = True
        else:
            log.debug("Checksum OK ({}). Decrypting...".format(self.file_hash.hexdigest()))

        cipher = AES.new(key.get(), AES.MODE_GCM, self.iv)
        try:
            output_decrypted: bytes = cipher.decrypt(encrypted_data)
        except ValueError as e:
            log.fatal("Decryption failed: {}."
                      "\n    This probably means your backup is corrupted.".format(e))
            raise e

        # Verify the authentication tag
        try:
            if is_multifile_backup:
                # In multifile backups, there is no checksum.
                # This means, the last 16 bytes of the files are not the checksum,
                # despite being called "checksum", but are the authentication tag.
                # Same way, "authentication tag" is not the tag, but the last
                # 16 bytes of the encrypted file.
                output_decrypted += cipher.decrypt(authentication_tag)
                cipher.verify(checksum)
            else:
                cipher.verify(authentication_tag)
        except ValueError as e:
            log.error("Authentication tag mismatch: {}."
                      "\n    This probably means your backup is corrupted.".format(e))

        return output_decrypted

    def encrypt(self, key: Key15, props: Props, decrypted: bytes) -> bytes:
        """Encrypts the database using the provided key"""
        from wa_crypt_tools.proto import C15_IV_pb2 as C15_IV
        cipher = C15_IV.C15_IV()
        cipher.IV = self.iv
        from wa_crypt_tools.proto import backup_prefix_pb2 as prefix
        from wa_crypt_tools.proto import key_type_pb2 as key_type
        prefix = prefix.BackupPrefix()
        prefix.key_type = key_type.Key_Type.HSM_CONTROLLED
        prefix.c15_iv.CopyFrom(cipher)

        prefix.info.CopyFrom(props.get_proto())
        prefix = prefix.SerializeToString()
        out = b''
        file_hash = md5()
        out += len(prefix).to_bytes(1, byteorder='big')
        file_hash.update(out)
        out += b'\x01'
        file_hash.update(b'\x01')
        out += prefix
        file_hash.update(prefix)
        cipher = AES.new(key.get(), AES.MODE_GCM, self.iv)
        encrypted_data, authentication_tag = cipher.encrypt_and_digest(decrypted)
        out += encrypted_data
        file_hash.update(encrypted_data)
        out += authentication_tag
        file_hash.update(authentication_tag)
        out += file_hash.digest()
        return out

    def get_iv(self) -> bytes:
        return self.iv



class C:
    # These constants are only used by the guessing logic.
    # zlib magic header is 78 01 (Low Compression).
    # The first two bytes of the decrypted data should be those,
    # in case of single file backup, or PK in case of multi file.
    ZLIB_HEADERS = [
        b'x\x01',
        b'PK'
    ]
    ZIP_HEADER = b'PK\x03\x04'
    # Size of bytes to test (number chosen arbitrarily, but values less than ~310 makes test_decompression fail)
    HEADER_SIZE = 384
    DEFAULT_DATA_OFFSET = 122
    DEFAULT_IV_OFFSET = 8

    # Encryption constants
    DEFAULT_APP_VERSION = "2.23.18.12"
    DEFAULT_JID_SUFFIX = "00"
    DEFAULT_BACKUP_VERSION = 0
    # The Props I got from a recent backup of mine
    DEFAULT_FEATURE_LIST = [5, 6, 7, 8, 9,
                            10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                            20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
                            30, 31, 32, 33, 35, 36, 37, 39]
    # Old backups might not have knowledge of the new features (in 2022 backups the max is 37)
    DEFAULT_MAX_FEATURE = 39

    # Constants for crypt12/14 key and db
    SUPPORTED_CIPHER_VERSION = b'\x00\x01'
    SUPPORTED_KEY_VERSIONS = [b'\x01', b'\x02', b'\x03']


import os
import zlib

from wa_crypt_tools.lib.db.db12 import Database12
from wa_crypt_tools.lib.db.db14 import Database14
from wa_crypt_tools.lib.db.db15 import Database15
from wa_crypt_tools.lib.db.dbfactory import DatabaseFactory
from wa_crypt_tools.lib.key.keyfactory import KeyFactory
from wa_crypt_tools.lib.props import Props
from hashlib import sha512

class TestDecryption:
    def test_decryption15(self):
        key = KeyFactory.new("tests/res/encrypted_backup.key")
        f = open("tests/res/msgstore.db.crypt15",'rb')
        db = DatabaseFactory.from_file(f)
        encrypted = f.read()
        decrypted_db = db.decrypt(key, encrypted)
        decrypted_db = zlib.decompress(decrypted_db)
        new_check = sha512(decrypted_db).digest()
        with open("tests/res/msgstore.db", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert new_check == orig_check
    
    def test_decryption14(self):
        key = KeyFactory.new("tests/res/key")
        f = open("tests/res/msgstore.db.crypt14",'rb')
        db = DatabaseFactory.from_file(f)
        encrypted = f.read()
        decrypted_db = db.decrypt(key, encrypted)
        decrypted_db = zlib.decompress(decrypted_db)
        new_check = sha512(decrypted_db).digest()
        with open("tests/res/msgstore.db", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert new_check == orig_check
        
    def test_decryption12(self):
        key = KeyFactory.new("tests/res/key")
        f = open("tests/res/msgstore.db.crypt12",'rb')
        db = DatabaseFactory.from_file(f)
        encrypted = f.read()
        decrypted_db = db.decrypt(key, encrypted)
        decrypted_db = zlib.decompress(decrypted_db)
        new_check = sha512(decrypted_db).digest()
        with open("tests/res/msgstore.db", 'rb') as f:
            orig_check = sha512(f.read()).digest()
        assert new_check == orig_check


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 27, 3, '', 'backup_prefix.proto')
_sym_db = _symbol_database.Default()
from . import C14_cipher_pb2 as C14__cipher__pb2
from . import C15_IV_pb2 as C15__IV__pb2
from . import key_type_pb2 as key__type__pb2
from . import backup_expiry_pb2 as backup__expiry__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13backup_prefix.proto\x1a\x10C14_cipher.proto\x1a\x0cC15_IV.proto\x1a\x0ekey_type.proto\x1a\x13backup_expiry.proto"\xa7\x01\n\x0cBackupPrefix\x12 \n\x08key_type\x18\x01 \x01(\x0e2\t.Key_TypeH\x01\x88\x01\x01\x12!\n\nc14_cipher\x18\x02 \x01(\x0b2\x0b.C14_cipherH\x00\x12\x19\n\x06c15_iv\x18\x03 \x01(\x0b2\x07.C15_IVH\x00\x12\x1b\n\x04info\x18\x04 \x01(\x0b2\r.BackupExpiryB\r\n\x0bcipher_infoB\x0b\n\t_key_typeb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'backup_prefix_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_BACKUPPREFIX']._serialized_start = 93
    _globals['_BACKUPPREFIX']._serialized_end = 260


import argparse
import hashlib
import os
import zlib
import logging

from wa_crypt_tools.lib.constants import C
from wa_crypt_tools.lib.db.db import Database
from wa_crypt_tools.lib.db.db12 import Database12
from wa_crypt_tools.lib.db.db14 import Database14
from wa_crypt_tools.lib.db.db15 import Database15
from wa_crypt_tools.lib.db.dbfactory import DatabaseFactory
from wa_crypt_tools.lib.key.key import Key
from wa_crypt_tools.lib.key.keyfactory import KeyFactory
from wa_crypt_tools.lib.logformat import CustomFormatter
from wa_crypt_tools.lib.props import Props

log = logging.getLogger(__name__)


def parsecmdline() -> argparse.Namespace:
    """Parses the command line arguments."""
    """Sets up the argument parser"""
    parser = argparse.ArgumentParser(description='Encrypts a file in Crypt14 or Crypt15 format.')
    parser.add_argument('keyfile', nargs='?', type=str, default="encrypted_backup.key",
                        help='The WhatsApp encrypted_backup key file or the hex encoded key. '
                             'Default: encrypted_backup.key')
    parser.add_argument('decrypted', nargs='?', type=argparse.FileType('rb'), default="msgstore.db",
                        help='The input file. Default: msgstore.db')
    parser.add_argument('encrypted', nargs='?', type=argparse.FileType('wb'), default="msgstore.db.crypt15",
                        help='The encrypted crypt15 or crypt14 file. Default: msgstore.db.crypt15')
    parser.add_argument('-f', '--force', action='store_true',
                        help='Makes errors non fatal. Default: false')
    parser.add_argument('-v', '--verbose', action='store_true', help='Prints all offsets and messages')
    parser.add_argument('--enable-features', type=int, nargs='*', default=C.DEFAULT_FEATURE_LIST,
                        help='Enables the specified features. ')
    parser.add_argument('--max-feature', type=int, default=39,
                        help='The max feature number, the older is the backup the lower should be the number. ')
    parser.add_argument('--multi-file', action='store_true',
                        help='Encrypts a multi-file backup (either stickers or wallpapers)')
    parser.add_argument('--type', type=int, choices=[12, 14, 15], default=15,
                        help='The type of encryption to use. Default: 15')
    parser.add_argument('--iv', type=str, help='The IV to use for crypt15 encryption. Default: random')
    parser.add_argument('--reference', type=argparse.FileType('rb'),
                        help='The reference file to use for crypt15 encryption. Highly recommended.')
    parser.add_argument('--noparse', action='store_true',
                        help='Do not parse the header of the reference file. Default: false')
    parser.add_argument('--wa-version', type=str, default=C.DEFAULT_APP_VERSION,
                        help='The WhatsApp version to use for crypt15 encryption. Default:' +
                             C.DEFAULT_APP_VERSION)
    parser.add_argument('--jid', type=str, default=C.DEFAULT_JID_SUFFIX,
                        help='The last 2 numbers of your phone number. Default: 00')
    parser.add_argument('--backup-version', type=int, default=C.DEFAULT_BACKUP_VERSION,
                        help='The backup version to use in the header of the encrypted file. Default: 0')
    parser.add_argument('--no-compress', action='store_true',
                        help='Do not compress the file. This will make the backup not working. Only used in development. Default: false')
    return parser.parse_args()


def main():
    """Main function"""
    # Parse the command line arguments
    args = parsecmdline()
    # set wa_crypt_tools l to debug
    log.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ch.setFormatter(CustomFormatter())
    log.addHandler(ch)
    log.warning("This script is in beta stage")

    # Read the key file
    key = KeyFactory.new(args.keyfile)
    # If specified, use the IV from the command line
    iv = None
    props = None
    if not args.reference:
        if args.iv:
            iv = bytes.fromhex(args.iv)
        # Create the props object from the command line arguments
        props = Props(wa_version=args.wa_version, jid=args.jid, max_feature=args.max_feature,
                      features=args.enable_features, backup_version=args.backup_version)
    else:
        reference = DatabaseFactory.from_file(args.reference)
        iv: bytes = reference.get_iv()
        props = reference.props
    data = args.decrypted.read()
    if args.type == 15:
        db = Database15(key=key, iv=iv)
    elif args.type == 14:
        db = Database14(key=key, iv=iv)
    else:
        db = Database12(key=key, iv=iv)
    if args.no_compress:
        encrypted = db.encrypt(key, props, data)
    else:
        compressed = zlib.compress(data, 1)
        encrypted = db.encrypt(key, props, compressed)
    args.encrypted.write(encrypted)
    # Close the files
    log.info("Done!")
    args.decrypted.close()
    args.encrypted.close()


if __name__ == '__main__':
    main()



from wa_crypt_tools.lib.utils import hexstring2bytes


class TestUtils:
    # Sample test to test the test infrastructure (!)
    def test_hexstring2bytes(self):
        assert hexstring2bytes("0"*64) == b'\x00' * 32
